"""
Groups

This module contains actions related to Active Directory Groups.
"""
import uuid
import functools
import logging

from typing import List, Dict, Optional, Union, Any


from ppa import ui
from active_directory import _ldaps, helpers, users, groups, computers, exceptions as sdk_exceptions
from playbook_tools import tables, process

from .exceptions import NoGroupFound, InvalidAttribute
from ._wrappers import (
    ldap_function,
    idempotent_create,
    idempotent_remove,
    process_user,
    process_group,
    process_computer,
    bool_to_search_scope,
)
from .records import Group, User, Computer, DomainController
from .users import input_table as users_input_table, output_table as users_output_table


GROUP_TABLE_HEADER = ["Common Name", "Description", "Type"]
GROUP_TABLE_FIELDS = ["cn", "description", "groupType"]


logger = logging.getLogger(__name__)


@ldap_function
def get_all(
    domain_controller: DomainController, search_base: Optional[str] = None, one_level: Optional[bool] = False
) -> List[Group]:
    """Get all groups.

!!! danger "This operation can be slow on large domains"

!!! question "Minimum Plugin Version: 6.1.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __search_base:__ optional start point for the search (see [here](../advanced_queries/#search-location) for more information)

    - __one_level:__ set to `true` to only search one level (see [here](../advanced_queries/#search-scopes) for more information)

??? success "Output"

    A list of [Group Dictionaries](../inputs_outputs/#group).

??? example
    ```YAML linenums="1"
    - active_directory.groups.get_all:
      load:
        domain_controller: domain_controller_info
      save: all_groups
    ```
"""
    logger.info(f"getting all groups using search base {search_base if search_base else 'default'}")
    return list(
        map(
            process_group,
            groups.get_all(domain_controller, search_base=search_base, search_scope=bool_to_search_scope(one_level)),
        )
    )


@ldap_function
def by_samaccountname(domain_controller: DomainController, sAMAccountName: str) -> Group:
    """Get a group by its 'sAMAccountName' attribute.

!!! question "Minimum Plugin Version: 1.3.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __sAMAccountName:__ The group's `sAMAccountName`

??? success "Output"
    A single [Group Dictionary](../inputs_outputs/#group).

??? example
    ```YAML linenums="1"
    - active_directory.groups.by_samaccountname:
        sAMAccountName: example.group
      load:
        domain_controller: domain_controller_info
      save: example_group
    ```
"""
    logger.info(f"getting group with sAMAccountName '{sAMAccountName}'")
    try:
        return process_group(groups.search(domain_controller, sAMAccountName=sAMAccountName)[0])
    except IndexError:
        raise NoGroupFound(f"No group found with sAMAccountName '{sAMAccountName}'.")


@ldap_function
def by_distinguishedname(domain_controller: DomainController, distinguishedName: str) -> Group:
    """Get a group by its 'distinguishedName' attribute.

!!! question "Minimum Plugin Version: 1.3.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ The group's `distinguishedName`

??? success "Output"
    A single [Group Dictionary](../inputs_outputs/#group).

??? example
    ```YAML linenums="1"
    - active_directory.groups.by_distinguishedname:
        distinguishedName: cn=Example Group,CN=Users,DC=Example,DC=Domain
      load:
        domain_controller: domain_controller_info
      save: example_group
    ```
"""
    logger.info(f"getting group with distinguishedName '{distinguishedName}'")
    try:
        return process_group(groups.search(domain_controller, distinguishedName=distinguishedName)[0])
    except IndexError:
        raise NoGroupFound(f"No group found with distinguishedName '{distinguishedName}'.")


@ldap_function
def get_interactive(
    domain_controller: DomainController,
    title: Optional[str] = None,
    exclude: Optional[Dict[str, Union[str, int]]] = None,
    search_base: Optional[str] = None,
) -> Group:
    """Start an interactive search for a group using its 'Common Name'.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __title:__ a title displayed to the Task Operator (defaults to `Get Security Group`)

    - __exclude:__ a dictionary containing LDAP attribute names & regular expressions to test their values

    - __search_base:__ optional start point for the search (see [here](../advanced_queries/#search-location) for more information)

??? success "Output"

    A single [Group Dictionary](../inputs_outputs/#group).

!!! tip "Automatic Wildcards"

    When using this action all provided search terms will have a wildcard appended.

??? example

    - Interactively searching for a group inside the builtin __Users__ CN

    - All groups whose `cn` contains __admin__ are excluded from the results

    ```YAML linenums="1"
    - active_directory.groups.get_interactive:
        search_base: CN=Users,DC=Example,DC=Domain,DC=Com
        exclude:
          cn: .*admin.*
      load:
        domain_controller: domain_controller_info
      save: group
    ```
    """
    logger.info("starting interactive group search")
    while True:
        try:
            return process_group(
                helpers.get_group(
                    domain_controller,
                    search_attribute="cn",
                    search_base=search_base,
                    auto_wildcard=True,
                    auto_select=False,
                    exclude=exclude,
                    title=title,
                )
            )
        except sdk_exceptions.ParameterError as e:
            if "has value *" in str(e):
                ui.output_error("Searching with only a wildcard is not supported for performance reasons, please try again.")
                continue
            raise e


@ldap_function
def get_members(
    domain_controller: DomainController, distinguishedName: str, nested: Optional[bool] = False
) -> List[Union[Group, User, Computer]]:
    """Get the members of a group.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __distinguishedName:__ the distinguished name of the group

    - __nested:__ set to `true` to include nested group members in the search

??? success "Output"
    A list of [User](../inputs_outputs/#user), [Group](../inputs_outputs/#group), & [Computer](../inputs_outputs/#computer) dictionaries.

??? warning
    Getting nested group memberships can be slow.

??? example
    Getting all members of _Example Group_, & saving them as a new variable called _group_members_.
    ```YAML linenums="1"
    - active_directory.groups.get_members:
        distinguishedName: "CN=Example Group,CN=Users,DC=Example,DC=Domain"
        nested: true
      load:
        domain_controller: domain_controller_info
      save: group_members
    ```
    """
    logger.info(f"getting all {'direct' if not nested else 'nested'} members group '{distinguishedName}'")
    record_generators = {
        "User": process_user,
        "Group": process_group,
        "Computer": process_computer,
    }
    try:
        return list(
            map(
                lambda m: record_generators[m._ppa_record_name](m),
                groups.members(domain_controller, distinguished_name=distinguishedName, nested=nested),
            )
        )
    except KeyError:
        raise KeyError("Group members contains a record that is not one of User, Group, Computer. Cannot continue.")


@ldap_function
def get_groups(domain_controller: DomainController, distinguishedName: str, nested: Optional[bool] = False) -> List[Group]:
    """Get groups from a group's membership list.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __distinguishedName:__ the distinguished name of the group

    - __nested:__ set to `true` to include nested group members in the search

??? success "Output"
    A list of [Group Dictionaries](../inputs_outputs/#group).

??? warning
    Getting nested group memberships can be slow.

??? example
    Getting all groups in _Example Group_, & saving them as a new variable called _group_list_.
    ```YAML linenums="1"
    - active_directory.groups.get_groups:
        distinguishedName: "CN=Example Group,CN=Users,DC=Example,DC=Domain"
        nested: true
      load:
        domain_controller: domain_controller_info
      save: group_list
    ```
    """
    logger.info(f"getting all {'direct' if not nested else 'nested'} groups from group '{distinguishedName}'")
    return list(
        map(
            process_group,
            list(
                filter(
                    lambda record: isinstance(record, groups.Group),
                    groups.members(domain_controller, distinguished_name=distinguishedName, nested=nested),
                )
            ),
        )
    )


@ldap_function
def get_users(domain_controller: DomainController, distinguishedName: str, nested: Optional[bool] = False) -> List[User]:
    """Get users from a group's membership list.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __distinguishedName:__ the distinguished name of the group

    - __nested:__ set to `true` to include nested group members in the search

??? success "Output"

    A list of [User Dictionaries](../inputs_outputs/#user).

??? warning
    Getting nested group memberships can be slow.

??? example
    Getting all users in _Example Group_, & saving them as a new variable called _user_list_.
    ```YAML linenums="1"
    - active_directory.groups.get_users:
        distinguishedName: "CN=Example Group,CN=Users,DC=Example,DC=Domain"
        nested: true
      load:
        domain_controller: domain_controller_info
      save: user_list
    ```
    """
    logger.info(f"getting all {'direct' if not nested else 'nested'} users from group '{distinguishedName}'")
    return list(
        map(
            process_user,
            list(
                filter(
                    lambda record: isinstance(record, users.User),
                    groups.members(domain_controller, distinguished_name=distinguishedName, nested=nested),
                )
            ),
        )
    )


@ldap_function
def get_computers(domain_controller: DomainController, distinguishedName: str, nested: Optional[bool] = False) -> List[Computer]:
    """Get computers from a group's membership list.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __distinguishedName:__ the distinguished name of the group

    - __nested:__ set to `true` to include nested group members in the search

??? success "Output"
    A list of [Computer Dictionaries](../inputs_outputs/#group).

??? warning
    Getting nested group memberships can be slow.

??? example
    - Getting all computers in _Example Group_

    - The results are saved as a new variable called `computer_list`

    ```YAML linenums="1"
    - active_directory.groups.get_computers:
        distinguishedName: "CN=Example Group,CN=Users,DC=Example,DC=Domain"
        nested: true
      load:
        domain_controller: domain_controller_info
      save: computer_list
    ```
    """
    logger.info(f"getting all {'direct' if not nested else 'nested'} computers from group '{distinguishedName}'")
    return list(
        map(
            process_computer,
            list(
                filter(
                    lambda record: isinstance(record, computers.Computer),
                    groups.members(domain_controller, distinguished_name=distinguishedName, nested=nested),
                )
            ),
        )
    )


@ldap_function
def search(
    domain_controller: DomainController,
    search_params: Dict[str, Any],
    search_base: Optional[str] = None,
    one_level: Optional[bool] = False,
) -> process.SearchResult:
    """Search for groups using LDAP attributes & values.

!!! question "Minimum Plugin Version: 6.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __search_params:__ a dictionary containing [group keys](../inputs_outputs#group) & values to use in the search

    - __search_base:__ optional start point for the search (see [here](../advanced_queries/#search-location) for more information)

    - __one_level:__ set to `true` to only search one level (see [here](../advanced_queries/#search-scopes) for more information)

??? success "Output"

    A [SearchResult](../../../searches/#search-result-dictionary) containing:

    - **total**: the number of [Groups](../inputs_outputs/#group) found

    - **all**: a list of [Groups](../inputs_outputs/#group) found

    - **first**: the first [Group](../inputs_outputs/#group) found

    - **last**: the last [Group](../inputs_outputs/#group) found

??? warning "Search Result Format"

    The search result format was updated in version 6.0.0 of this plugin.

    It is not compatible with playbooks written against previous versions.

    See [here](/searches/#using-search-results) for more information.

??? example

    Searching for all groups in the __Users__ CN whose `cn` starts with __admin*__:

    ```YAML linenums="1"
    - active_directory.groups.search:
        search_params:
          cn: admin*
        search_base: CN=Users,DC=Example,DC=Domain,DC=Com
      load:
        domain_controller: domain_controller_info
      save: groups
    ```

!!! tip

    You can use `*` as a wildcard at the end of search values.
"""
    logger.info(
        f"searching for groups in search base {search_base if search_base else 'default'} using the following parameters: {search_params}"
    )
    return process.search_result(
        list(
            map(
                process_group,
                groups.search(
                    domain_controller, search_base=search_base, search_scope=bool_to_search_scope(one_level), **search_params
                ),
            )
        )
    )


@ldap_function
@idempotent_create
def add_member(domain_controller: DomainController, group_distinguishedName: str, member_distinguishedName: str) -> None:
    """Add a member to a group.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __group_distinguishedName:__ the distinguishedName of the group

    - __member_distinguishedName:__ the distinguishedName of the new member

??? success "Output"
    Nothing is outputted by this action.

??? example
    Adding a user to a group.

    - The user & group are searched for using the _sAMAccountName_ attribute

    - They are saved as new variables `user` & `group`

    - The required _distinguishedName_ values are supplied using fields from the `user` & `group` variables

    ```YAML linenums="1"
    - active_directory.groups.search:
        sAMAccountName: example.group
      load:
        domain_controller: domain_controller_info
      save: group

    - active_directory.users.search:
        sAMAccountName: example.user
      load:
        domain_controller: domain_controller_info
      save: user

    - active_directory.groups.add_member:
      load:
        group_distinguishedName: group.distinguishedName
        member_distinguishedName: member.distinguishedName
        domain_controller: domain_controller_info
    ```
"""
    try:
        return groups.add_member(
            domain_controller, distinguished_name=group_distinguishedName, member_distinguished_name=member_distinguishedName,
        )
    except sdk_exceptions.OperationFailed as e:
        logger.exception(e)
        ui.output_error(str(e))
        raise e


@ldap_function
@idempotent_remove
def remove_member(domain_controller: DomainController, group_distinguishedName: str, member_distinguishedName: str) -> None:
    """Remove a member from a group.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __group_distinguishedName:__ the _distinguishedName_ of the group

    - __member_distinguishedName:__ the _distinguishedName_ of the member to remove

??? success "Output"
    Nothing is outputted by this action.

??? example

    Removing user _Example User_ from the group _Example Group_.

    ```YAML linenums="1"
    - active_directory.groups.remove_member:
        group_distinguishedName: "CN=Example Group,CN=Users,DC=Example,DC=Domain"
        member_distinguishedName: "CN=Example User,CN=Users,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
    ```
    """
    return groups.remove_member(
        domain_controller, distinguished_name=group_distinguishedName, member_distinguished_name=member_distinguishedName,
    )


@ldap_function
@idempotent_create
def create(
    domain_controller: DomainController,
    sAMAccountName: str,
    name: str,
    distinguishedName: str,
    group_type: str = "global",
    extra_params: Optional[Dict[str, Any]] = None,
) -> None:
    """Create a new security group.

!!! question "Minimum Plugin Version: 3.3.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __sAMAccountName:__ the new group sAMAccountName

    - __name:__ the new group name

    - __group_type:__ `domain local`, `universal`, or `global` (defaults to `global`)

    - __distinguishedName:__ the new group distinguishedName

    - __extra_params:__ a dictionary containing any extra LDAP attributes & values for the group

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - active_directory.groups.create:
        sAMAccountName: example.group
        distinguishedName: "CN=Example Group,CN=Users,DC=Example,DC=Domain"
        group_type: domain local
        name: example.group
        extra_params:
          info: An example Security Group
          cn: Example Group
      load:
        domain_controller: domain_controller_info
    ```
    """
    valid_group_types = {"domain local", "global", "universal"}
    if group_type not in valid_group_types:
        message = "Supplied group type '{group_type}' must be one of '{supported_types}'.".format(
            group_type=group_type, supported_types=", ".join(valid_group_types)
        )
        ui.output_error(message)
        raise ValueError(message)

    logger.info(f"creating group '{sAMAccountName}' at '{distinguishedName}' with the following extra parameters: {extra_params}")
    try:
        return groups.create(
            domain_controller,
            distinguished_name=distinguishedName,
            name=name,
            sAMAccountName=sAMAccountName,
            group_type=group_type.replace(" ", "_"),
            **extra_params or {},
        )
    except _ldaps.ldap.OBJECT_CLASS_VIOLATION as e:
        logger.error(e)
        if processed_exception := _ldaps._process_ldap_exception(e):
            if all([processed_exception.error == "0000207D", processed_exception.data == "131213"]):
                logger.error(f"processed ldap exception: {processed_exception}")
                message = "One or more of the following is not a supported {} group attribute: {}.".format(
                    group_type, ", ".join([f"'{key}'" for key in list(extra_params.keys())])
                )
                ui.output_error(message)
                raise InvalidAttribute(message)
        raise e


@ldap_function
@idempotent_remove
def delete(domain_controller: DomainController, distinguishedName: str) -> None:
    """Delete a group.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the group to delete

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - active_directory.groups.delete:
        distinguishedName: "CN=Example Group,CN=Users,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
    ```
    """
    logger.info(f"deleting group with distinguished name {distinguishedName}")
    return groups.delete(domain_controller, distinguished_name=distinguishedName)


@ldap_function
def move(domain_controller: DomainController, distinguishedName: str, parent_distinguishedName: str) -> None:
    """Move a group to a different OU or CN.

!!! question "Minimum Plugin Version: 1.5.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the group to move

    - __parent_distinguishedName:__ the distinguishedName of the OU or CN to move the group to

??? success "Output"

    Nothing is outputted by this action.

??? example

    Moving a group from the 'Users' CN to the 'Staff Groups' OU:

    ```YAML linenums="1"
    - active_directory.groups.move:
        distinguishedName: "CN=Example Group,CN=Users,DC=Example,DC=Domain"
        parent_distinguishedName: "OU=Staff Groups,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
    ```
    """
    return groups.move(domain_controller, distinguished_name=distinguishedName, new_parent=parent_distinguishedName)


@ldap_function
def delete_attribute(domain_controller: DomainController, distinguishedName: str, name: str, value: Union[str, int]) -> None:
    """Delete a certain value from a group attribute.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the group

    - __name:__ the name of the LDAP attribute

    - __value:__ the value to delete

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - active_directory.groups.get_interactive:
        cn: Example Group
      load:
        domain_controller: domain_controller_info
      save: group

    - active_directory.groups.delete_attribute:
        name: info
      load:
        distinguishedName: group.distinguishedName
        value: group.info
        domain_controller: domain_controller_info
    ```
    """
    return groups.delete_attribute(domain_controller, distinguished_name=distinguishedName, name=name, value=value)


@ldap_function
def clear_attribute(domain_controller: DomainController, distinguishedName: str, name: str) -> None:
    """Clear a particular LDAP attribute on a group.

Works for single-valued & multi-valued string LDAP attributes.

!!! question "Minimum Plugin Version: 3.1.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the group

    - __name:__ the name of the LDAP attribute

??? success "Output"

    Nothing is outputted by this action.

??? example

    Searching for a group & clearing its __info__ field:

    ```YAML linenums="1"
    - active_directory.groups.by_samaccountname:
        sAMAccountName: Support Staff
      load:
        domain_controller: domain_controller
      save: group

    - active_directory.groups.clear_attribute:
        name: info
      load:
        distinguishedName: group.distinguishedName
        domain_controller: domain_controller
    ```
"""
    existing_value = by_distinguishedname(domain_controller, distinguishedName)[name]

    if existing_value is None:
        # Nothing to do as the attribute has no value.
        return

    delete_func = functools.partial(delete_attribute, domain_controller, distinguishedName, name=name)
    list(map(lambda v: delete_func(value=v), existing_value)) if isinstance(existing_value, list) else delete_func(
        value=existing_value
    )


@ldap_function
def set_attribute(domain_controller: DomainController, distinguishedName: str, name: str, value: Union[str, int]) -> None:
    """Set a group attribute value.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the group

    - __name:__ the name of the LDAP attribute

    - __value:__ the value to set

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - active_directory.groups.get_interactive:
        cn: Example Group
      load:
        domain_controller: domain_controller_info
      save: group

    - active_directory.groups.set_attribute:
        name: info
      load:
        distinguishedName: group.distinguishedName
        value: Example Info Value
        domain_controller: domain_controller_info
    ```
    """
    return groups.set_attribute(domain_controller, distinguished_name=distinguishedName, name=name, value=value)


@ldap_function
def replace_attribute(domain_controller: DomainController, distinguishedName: str, name: str, value: Union[str, int]) -> None:
    """Replace a group attribute value.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"
    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the group

    - __name:__ the name of the LDAP attribute

    - __value:__ the value to set

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - active_directory.groups.get_interactive:
        cn: Example Group
      load:
        domain_controller: domain_controller_info
      save: group

    - active_directory.groups.replace_attribute:
        name: info
        value: Example Info Value
      load:
        distinguishedName: group.distinguishedName
        domain_controller: domain_controller_info
    ```
    """
    return groups.replace_attribute(domain_controller, distinguished_name=distinguishedName, name=name, value=value)


@ldap_function
def review_users(domain_controller: DomainController, distinguishedName: str, remove: Optional[bool] = False) -> Dict[str, Any]:
    """Review all users with a direct membership to the supplied group.

If a user should be in the group, their membership is confirmed by selecting them in the table.

If a user should _not_ be in the group, they should be left unselected in the table.

This action can also remove unselected users from the group (see below for more).

!!! question "Minimum Plugin Version: 5.1.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the group

    - __remove:__ when `true` PPA will remove unselected users from the group (defaults to `false`)

??? tip "Removing Unselected Users"

    By default PPA will _not remove_ unselected users from the group.

    This default setting can be used to collect _proposed_ changes without actually performing them.

    To remove unselected users in this action, set the __remove__ input to `true`.

    PPA will always ask for confirmation before any group memberships are modified.

??? success "Output"

    A dictionary containing the following keys & values:

    - __users_to_keep:__ a list of users who _should_ be in the group

    - __users_to_remove:__ a list of users who _should not_ be in the group

    If __remove__ is `true`, the users in __users_to_remove__ will have been removed from the group.

    The users in both lists will be [User Dictionaries](../inputs_outputs/#user).

??? example

    {% raw %}
    ```YAML linenums="1"
    - active_directory.groups.review_users:
        remove: true
      load:
        distinguishedName: group.distinguishedName
        domain_controller: domain_controller_info
      save: review

    - ppa.ui.output_info:
        text: PPA removed {{ review.users_to_remove | length }} users from {{ group.cn }}
    ```
    {% endraw %}
    """
    logger.info(f"reviewing users in group {distinguishedName}")
    logger.info(f"remove is {remove}")
    review_template = """### Review Complete

__Group:__ {group_name}

__Valid Users:__ {keep_count}

__Invalid Users:__ {remove_count}

{message}"""

    group = by_distinguishedname(domain_controller, distinguishedName)
    group_name = group.get("cn", group.get("sAMAccountName"))
    all_users = get_users(domain_controller, distinguishedName)

    if not all_users:
        ui.output_info(f"There are no users to review in {group_name}.")
        ui.input_accept("OK")
        return {"users_to_remove": [], "users_to_keep": []}

    ui.output_markdown(
        f"""### Review Group Memberships

__Group:__ _{group_name}_

Please review the users currently in the group.

- Select a user to confirm they _should_ be in the group

- Leave a user unselected to confirm they _should not_ be in the group

- You will be asked to confirm your choice if any users are left unselected

__Note:__ PPA will {'__remove__' if remove else '__not remove__'} unselected users from the group.
"""
    )
    ui.input_accept("OK")

    while True:
        users_to_keep = users_input_table(f"Review users for {group_name}", users=all_users)
        users_to_remove = []

        # No users were selected to be kept in the group.
        if not users_to_keep:
            ui.output_error("You have not selected any users.")
            # Should they all be removed?
            if ui.input_confirm(f"Should all the users above be removed from {group_name}?", confirm="Yes", cancel="No"):
                users_to_remove = all_users
                break
            else:
                # Choose again?
                if ui.input_confirm("Select users again?", confirm="Yes", cancel="No"):
                    # Yes.
                    continue
                # No - do nothing.
                ui.output_info(f"No changes submitted for {group_name}.")
                return {"users_to_remove": [], "users_to_keep": all_users}

        # All users should be kept in the group.
        elif len(users_to_keep) == len(all_users):
            ui.output_info("You selected all users.")
            if ui.input_confirm(f"Should all the users above be in {group_name}?", confirm="Yes", cancel="No"):
                return {"users_to_remove": [], "users_to_keep": all_users}
            # Choose again?
            if ui.input_confirm("Select users again?", confirm="Yes", cancel="No"):
                # Yes.
                continue
            # No - do nothing.
            ui.output_info(f"No changes submitted for {group_name}.")
            return {"users_to_remove": [], "users_to_keep": all_users}

        # Some users were selected to be kept in the group.
        else:
            users_to_remove = [u for u in all_users if u not in users_to_keep]

            # Ask for confirmation before removing.
            users_output_table("Unselected users", users=users_to_remove)
            if ui.input_confirm(
                f"Remove these users from {group_name}?" if remove else f"Should these users be removed from {group_name}?",
                confirm="Yes",
                cancel="No",
            ):
                # Changes were confirmed.
                break

            # Change not confirmed! Select again?
            if ui.input_confirm("Select users again?", confirm="Yes", cancel="No"):
                # Yes.
                continue

            # No - do nothing.
            ui.output_info(f"No changes submitted for {group_name}.")
            return {"users_to_remove": [], "users_to_keep": all_users}

    review = {"users_to_keep": users_to_keep, "users_to_remove": users_to_remove}
    keep_count = len(users_to_keep)
    remove_count = len(users_to_remove)

    if not remove:
        ui.output_markdown(
            review_template.format(
                group_name=group_name,
                keep_count=keep_count,
                remove_count=remove_count,
                message="No group memberships have been modified.",
            )
        )
        logger.info("returning review without removing any memberships")
        return review

    logger.info(f"removing {remove_count} invalid members")
    ui.output_progress(
        f"Removing {remove_count} users from {group_name}...", percent=None, element_id=f"removing_users_{group_name}",
    )
    for i, user in enumerate(users_to_remove, 1):
        remove_member(
            domain_controller, group_distinguishedName=distinguishedName, member_distinguishedName=user["distinguishedName"]
        )
    ui.output_progress(
        f"Removed {remove_count} user(s) from {group_name}", percent=100, element_id=f"removing_users_{group_name}"
    )

    ui.output_markdown(
        review_template.format(
            group_name=group_name,
            keep_count=keep_count,
            remove_count=remove_count,
            message="All invalid users have been removed from the group.",
        )
    )

    return review


def output_table(text: str, groups: Union[Group, List[Group]]) -> None:
    """Display groups in a table.

The table will have the following columns:

- Common Name
- Description
- Type

!!! warning "Deprecation Warning"

    This action will soon be deprecated in favour of [display](#display).

!!! question "Minimum Plugin Version: 5.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __groups:__ a single or list of [Group Dictionaries](../inputs_outputs/#group)

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - active_directory.groups.search:
        cn: Operations
      load:
        domain_controller: domain_controller_info
      save: operations_groups

    - active_directory.groups.output_table:
        text: Operations Groups
      load:
        groups: operations_groups
    ```
    """
    groups = [groups] if not isinstance(groups, list) else groups
    ui.output_table(
        text,
        header=GROUP_TABLE_HEADER,
        data=[
            [str(group.get(field)) if group.get(field) is not None else "" for field in GROUP_TABLE_FIELDS] for group in groups
        ],
    )


def display(text: str, groups: Union[Group, List[Group]]) -> None:
    """Display groups in a table.

The table will have the following columns:

- Common Name
- Description
- Type

!!! question "Minimum Plugin Version: 6.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __groups:__ any number of [Groups](../inputs_outputs/#group)

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - active_directory.groups.search:
        cn: Operations
      load:
        domain_controller: domain_controller_info
      save: operations_groups

    - active_directory.groups.display:
        text: Operations Groups
      load:
        groups: operations_groups
    ```
    """
    tables.output_table(text, header=GROUP_TABLE_HEADER, extract=GROUP_TABLE_FIELDS, records=groups)


def output_tabbed_table(text: str, tabs: Dict[str, List[Group]]) -> None:
    """Display a groups table with multiple tabs.

This action can be used to display group lists from multiple Active Directories.

The table will have the following columns:

- Common Name
- Description
- Type

!!! warning "Deprecation Warning"

    This action will soon be deprecated in favour of [display_tabbed](#display_tabbed).

!!! question "Minimum Plugin Version: 5.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __tabs:__ a dictionary where each key is a tab name & each value is a list of [Group](../inputs_outputs/#group) dictionaries

??? success "Output"

    Nothing is outputted by this action.

??? example

    Auditing groups from 2 domains & presenting them in a tabbed table:

    ```YAML linenums="1"
    - set:
        name: tabbed_table
        value: {}

    - active_directory.groups.get_all:
      load:
        domain_controller: production_domain
      save: production_groups

    - active_directory.groups.get_all:
      load:
        domain_controller: uat_domain
      save: uat_groups

    - ppa_tools.dictionaries.insert:
        name: Production Active Directory
      load:
        value: production_groups
        dictionary: tabbed_table
      save: tabbed_table

    - ppa_tools.dictionaries.insert:
        name: UAT Active Directory
      load:
        value: uat_groups
        dictionary: tabbed_table
      save: tabbed_table

    - active_directory.groups.output_tabbed_table:
        text: Active Directory Groups
      load:
        tabs: tabbed_table
    ```
    """
    ui.output_table(
        text,
        header=GROUP_TABLE_HEADER,
        data={
            key: [
                [str(group.get(field)) if group.get(field) is not None else "" for field in GROUP_TABLE_FIELDS]
                for group in groups
            ]
            for key, groups in tabs.items()
        },
    )


def display_tabbed(text: str, tabs: Dict[str, List[Group]]) -> None:
    """Display a groups table with multiple tabs.

This action can be used to display group lists from multiple Active Directories.

The table will have the following columns:

- Common Name
- Description
- Type

!!! question "Minimum Plugin Version: 6.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __tabs:__ a dictionary where each key is a tab name & each value is any number of [Groups](../inputs_outputs/#group)

??? success "Output"

    Nothing is outputted by this action.

??? example

    Auditing groups from 2 domains & presenting them in a tabbed table:

    ```YAML linenums="1"
    - set:
        name: tabs
        value: {}

    - active_directory.groups.get_all:
      load:
        domain_controller: production_domain
      save: production_groups

    - active_directory.groups.get_all:
      load:
        domain_controller: uat_domain
      save: uat_groups

    - ppa_tools.dictionaries.insert:
        name: Production Active Directory
      load:
        value: production_groups
        dictionary: tabs
      save: tabs

    - ppa_tools.dictionaries.insert:
        name: UAT Active Directory
      load:
        value: uat_groups
        dictionary: tabs
      save: tabs

    - active_directory.groups.display_tabbed:
        text: Active Directory Groups
      load:
        tabs: tabs
    ```
    """
    tables.output_tabbed_table(
        text, header=GROUP_TABLE_HEADER, extract=GROUP_TABLE_FIELDS, tabs=tabs,
    )


def output_custom_table(text: str, header: List[str], attributes: List[str], groups: List[Group]) -> None:
    """Display groups in a table using custom headers & LDAP attributes.

!!! question "Minimum Plugin Version: 5.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __header:__ a list of column names

    - __attributes:__ a list of LDAP attributes (one for each column name in the `header`)

    - __groups:__ a single or list of [Group Dictionaries](../inputs_outputs/#group) to display in the table

??? success "Output"
    Nothing is outputted by this action.

??? example
    Showing `cn` & _groupType_ for groups in the variable `test_groups`.

    ```YAML linenums="1"
    - active_directory.groups.output_custom_table:
        text: Custom Groups Table
        header:
          - Common Name
          - Group Type
        attributes:
          - cn
          - groupType
      load:
        groups: test_groups
    ```
    """
    groups = [groups] if not isinstance(groups, list) else groups
    ui.output_table(
        text,
        header=header,
        data=[[str(group.get(attr)) if group.get(attr) is not None else "" for attr in attributes] for group in groups],
    )


def input_table(
    text: str, groups: Union[Group, List[Group]], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> List[Group]:
    """Display groups in a table, & allow the task operator to make a selection.

The table will have the following columns:

- Common Name
- Description
- Type

!!! warning "Deprecation Warning"

    This action will soon be deprecated in favour of [select](#select).

!!! question "Minimum Plugin Version: 5.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __groups:__ a single or list of [Group Dictionaries](../inputs_outputs/#group) to display in the table

    - __minimum:__ the minimum number of acceptable selections

    - __maximum:__ the maximum number of acceptable selections

??? success "Output"
    A list of [Group Dictionaries](../inputs_outputs/#group).

??? tip

    - If neither a _minimum_ or _maximum_ is provided, the task operator will be able to submit 0 selections.
    - If _minimum_ or _maximum_ are provided, the operation will repeat until the task operator makes a valid number
        of selections.

??? example
    - Getting all groups whose common names start with _Test_

    - Saving the results as a new variable called `test_groups`

    - Using this action to show the `test_groups` in a table, requiring the task operator selects at least 1

    - The selection is saved as a new variable called `selected_groups`

    ```YAML linenums="1"
    - active_directory.groups.search:
        cn: Test*
      load:
        domain_controller: domain_controller_info
      save: test_groups

    - active_directory.groups.input_table:
        text: "Please Select >= 1 Group(s)"
        minimum: 1
      load:
        groups: test_groups
      save: selected_groups
    ```
    """
    groups = [groups] if not isinstance(groups, list) else groups
    table_element_id = str(uuid.uuid4())
    error_element_id = str(uuid.uuid4())
    while True:
        indices = ui.input_table(
            text,
            header=GROUP_TABLE_HEADER,
            data=[
                [str(group.get(field)) if group.get(field) is not None else "" for field in GROUP_TABLE_FIELDS]
                for group in groups
            ],
            element_id=table_element_id,
        )
        count = len(indices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} record(s)", element_id=error_element_id)
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} record(s)", element_id=error_element_id)
            continue
        break

    return [groups[i] for i in indices]


def select(
    text: str, groups: Union[Group, List[Group]], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> tables.TableSelection:
    """Display groups in a table & prompt the task operator to make a selection.

The table will have the following columns:

- Common Name
- Description
- Type

!!! question "Minimum Plugin Version: 6.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __groups:__ any number of [Groups](../inputs_outputs/#group) to display in the table

    - __minimum:__ the minimum number of acceptable selections

    - __maximum:__ the maximum number of acceptable selections

??? success "Output"

    A [Selection](../../../tables/#selection) containing:

    - **total**: the number of selected [Groups](../inputs_outputs/#group)

    - **all**: a list of selected [Groups](../inputs_outputs/#group)

    - **first**: the first selected [Group](../inputs_outputs/#group)

??? example

    ```YAML linenums="1"
    - active_directory.groups.get_all:
      load:
        domain_controller: domain_controller
      save: groups

    - active_directory.groups.select:
        text: Select Groups
      load:
        groups: groups
      save: selection
    ```
    """
    return tables.input_table(
        text, header=GROUP_TABLE_HEADER, extract=GROUP_TABLE_FIELDS, records=groups, minimum=minimum, maximum=maximum
    )


def select_one(text: str, groups: Union[Group, List[Group]]) -> Group:
    """Display groups in a table & prompt the task operator to select one.

The table will have the following columns:

- Common Name
- Description
- Type

!!! question "Minimum Plugin Version: 6.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __groups:__ any number of [Groups](../inputs_outputs/#group)

??? success "Output"

    A single [Group](../inputs_outputs/#group).

??? example

    ```YAML linenums="1"
    - active_directory.groups.get_all:
      load:
        domain_controller: domain_controller
      save: groups

    - active_directory.groups.select_one:
        text: Select a Group
      load:
        groups: groups
      save: group
    ```
    """
    return select(text, groups, minimum=1, maximum=1)["first"]


@ldap_function
def exists(domain_controller: DomainController, **search_params: Union[str, int, None]) -> bool:
    """Search for groups using LDAP attributes & values to identify if any were found.

!!! question "Minimum Plugin Version: 1.0.0"


??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __search_params:__ a dictionary containing [group keys](../inputs_outputs#group) & values to use in the search


??? success "Output"

    A boolean is outputted by this action

    - `true` if one or more groups are found

    - `false` if no groups are found


??? example

    Searching using a unique attribute:

    ```YAML linenums="1"
    - active_directory.groups.exists:
        distinguishedName: "CN=Group Name,OU=Example,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
      save: item_exists
    ```

    Searching using a non-unique attribute:

    ```YAML linenums="1"
    - active_directory.groups.exists:
        cn: "My Group*"
      load:
        domain_controller: domain_controller_info
      save: item_exists
    ```

!!! tip "Wildcard Searching"

    It is possible to use `*` as a wildcard at the __end__ of search values, but this can make the search slow.

    """
    return bool(groups.search(domain_controller, **search_params))
